package alrosine.ean.upc.impl;

import alrosine.json.Encoding;
import alrosine.json.JsonGenerator;
import alrosine.json.JsonObject;
import alrosine.json.exception.JsonDuplicateNameException;
import alrosine.json.exception.JsonInvalidFormatException;
import alrosine.ean.upc.UpcResponse;

import java.io.IOException;
import java.io.InputStream;
import java.text.ParseException;
import java.text.SimpleDateFormat;

/**
 * JSON implementation of the UpcQueryParser interface
 */
public class UpcQueryJsonParser implements UpcQueryParser {

	private static String STATUS_ELEMENT_NAME = "status";
	private static String PRODUCT_ELEMENT_NAME = "product";
	private static String COMPANY_ELEMENT_NAME = "company";

	private boolean autoRetrieveGraphics = false;
	private UpcResponseImpl response;

	public UpcQueryJsonParser() {
	}

	public UpcQueryJsonParser(boolean autoRetrieveGraphics) {
		this.autoRetrieveGraphics = autoRetrieveGraphics;
	}

	@Override
	public UpcQueryServiceMode getUpcQueryServiceMode() {
		return UpcQueryServiceMode.Json;
	}

	@Override
	public UpcResponse getUpcResponse(InputStream inputStream) {
		try {
			byte[] inputData = new byte[inputStream.available()];
			inputStream.read(inputData);
			parse(decodeInput(inputData));
			return this.response;
		} catch(IOException e) {
			e.printStackTrace();
			throw new RuntimeException("Attempt to parse InputStream failed.", e);
		} catch(JsonDuplicateNameException e) {
			e.printStackTrace();
			throw new RuntimeException("Received Json data stream contained a duplicate named object.", e);
		} catch(JsonInvalidFormatException e) {
			e.printStackTrace();
			throw new RuntimeException("Received Json data stream contained invalid data.", e);
		}
	}

	private byte[] decodeInput(byte[] inputBytes) {
		String inputString = new String(inputBytes);
		for(Encoding cc : Encoding.values()) {
			if (!cc.getEncodedValue().equals(cc.getDecodedValue())) {
				int idx = inputString.indexOf(cc.getEncodedValue());
				while(idx > -1) {
					if (idx == 0) {
						inputString = cc.getDecodedValue() + inputString.substring(idx + 2);
					} else {
						inputString = inputString.substring(0, idx) + cc.getDecodedValue() + inputString.substring(idx + cc.getEncodedValue().length());
					}
					idx = inputString.indexOf(cc.getEncodedValue(), idx + cc.getDecodedValue().length());
				}
			}
		}
		return inputString.getBytes();
	}

	private void parse(byte[] inputData) {

		JsonObject jsonObject = JsonGenerator.generate(inputData);

		this.response = new UpcResponseImpl();
		UpcProductImpl product = new UpcProductImpl();
		UpcCompanyImpl company = new UpcCompanyImpl();
		this.response.setProduct(product);
		this.response.setCompany(company);

		JsonObject statusObject = (JsonObject) jsonObject.getObject(STATUS_ELEMENT_NAME);

		response.setRequestType(statusObject.getObject("message").toString());
		response.setResponseVersion(Integer.valueOf(statusObject.getObject("version").toString()));
		response.setReturnCode(Integer.valueOf(statusObject.getObject("code").toString()));
		response.setUpc(statusObject.getObject("find").toString());

		JsonObject productObject = (JsonObject) jsonObject.getObject(PRODUCT_ELEMENT_NAME);

		SimpleDateFormat formatter = new SimpleDateFormat("yyyy-MM-dd HH24:mm:ss");
		try {
			product.setLastModified(formatter.parse(productObject.getObject("modified").toString()));
		} catch(ParseException e) {
			//Do nothing...
		}

		product.setEan13(productObject.getObject("ean13").toString());
		product.setUpcA(productObject.getObject("upca").toString());
		product.setUpcE(productObject.getObject("upce").toString());
		product.setIsbn(productObject.getObject("isbn10").toString());
		product.setName(productObject.getObject("product").toString());
		product.setDescription(productObject.getObject("description").toString());
		product.setCategoryCode(productObject.getObject("category_no").toString());
		product.setCategoryName(productObject.getObject("category_text").toString());
		product.setCompanyUrl(productObject.getObject("url").toString());
		product.setLongDescription(Boolean.valueOf(productObject.getObject("has_long_desc").toString()));
		product.setProductImageUrl(productObject.getObject("image").toString());
		if (autoRetrieveGraphics) {
			product.setProductImage(UpcUtility.retrieveImage(product.getProductImageUrl()));
		}
		product.setBarcodeUrl(productObject.getObject("barcode").toString());
		if (autoRetrieveGraphics) {
			product.setBarcodeImage(UpcUtility.retrieveImage(product.getProductImageUrl()));
		}
		product.setLocked(Boolean.valueOf(productObject.getObject("locked").toString()));




		JsonObject companyObject = (JsonObject) jsonObject.getObject(COMPANY_ELEMENT_NAME);

		company.setName(companyObject.getObject("name").toString());
		company.setLogoUrl(companyObject.getObject("logo").toString());
		if (autoRetrieveGraphics) {
			company.setLogoImage(UpcUtility.retrieveImage(company.getLogoUrl()));
		}
		company.setUrl(companyObject.getObject("url").toString());
		company.setAddress(companyObject.getObject("address").toString());
		company.setTelephone(companyObject.getObject("phone").toString());
		company.setLocked(Boolean.valueOf(companyObject.getObject("locked").toString()));


	}

}
