package alrosine.ean.upc.impl;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;
import alrosine.ean.upc.UpcResponse;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import java.io.IOException;
import java.io.InputStream;
import java.text.ParseException;
import java.text.SimpleDateFormat;

/**
 * XML implementation of the UpcQueryParser interface
 */
public class UpcQueryXmlParser extends DefaultHandler implements UpcQueryParser {

	private static String FEED_ELEMENT_NAME = "feed";
	private static String STATUS_ELEMENT_NAME = "status";
	private static String PRODUCT_ELEMENT_NAME = "product";
	private static String COMPANY_ELEMENT_NAME = "company";

	private static String VERSION_ELEMENT_NAME = "version";
	private static String CODE_ELEMENT_NAME = "code";
	private static String MESSAGE_ELEMENT_NAME = "message";
	private static String FIND_ELEMENT_NAME = "find";

	private static String MODIFIED_ELEMENT_NAME = "modified";
	private static String EAN13_ELEMENT_NAME = "ean13";
	private static String UPCA_ELEMENT_NAME = "upca";
	private static String UPCE_ELEMENT_NAME = "upce";
	private static String ISBN_ELEMENT_NAME = "isbn10";
	private static String DESCRIPTION_ELEMENT_NAME = "description";
	private static String CATEGORY_NUMBER_ELEMENT_NAME = "category_no";
	private static String CATEGORY_TEXT_ELEMENT_NAME = "category_text";
	private static String URL_ELEMENT_NAME = "url";
	private static String LONG_DESCRIPTION_ELEMENT_NAME = "has_long_desc";
	private static String IMAGE_ELEMENT_NAME = "image";
	private static String BARCODE_ELEMENT_NAME = "barcode";
	private static String LOCKED_ELEMENT_NAME = "locked";

	private static String NAME_ELEMENT_NAME = "name";
	private static String LOGO_ELEMENT_NAME = "logo";
	private static String ADDRESS_ELEMENT_NAME = "address";
	private static String PHONE_ELEMENT_NAME = "phone";

	private static SimpleDateFormat dateFormatter = new SimpleDateFormat("yyyy-MM-dd HH24:mm:ss");

	private boolean autoRetrieveGraphics = false;

	private int feedCount = 0;
	private int statusCount = 0;
	private int productCount = 0;
	private int companyCount = 0;

	private boolean inFeed = false;
	private boolean inStatus = false;
	private boolean inProduct = false;
	private boolean inCompany = false;

	private String currentName = "";
	private UpcResponseImpl response;
	private UpcProductImpl product;
	private UpcCompanyImpl company;

	public UpcQueryXmlParser() {
	}

	public UpcQueryXmlParser(boolean autoRetrieveGraphics) {
		this.autoRetrieveGraphics = autoRetrieveGraphics;
	}

	@Override
	public UpcQueryServiceMode getUpcQueryServiceMode() {
		return UpcQueryServiceMode.Xml;
	}

	@Override
	public UpcResponse getUpcResponse(InputStream inputStream) {
		try {
			SAXParserFactory factory = SAXParserFactory.newInstance();
			SAXParser saxParser = factory.newSAXParser();
			saxParser.parse(inputStream, this);
			return this.response;
		} catch(ParserConfigurationException e) {
			e.printStackTrace();
			throw new RuntimeException("Attempt to parse InputStream failed.", e);
		} catch(SAXException e) {
			e.printStackTrace();
			throw new RuntimeException("Attempt to parse InputStream failed.", e);
		} catch(IOException e) {
			e.printStackTrace();
			throw new RuntimeException("Attempt to parse InputStream failed.", e);
		}
	}

	public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {

		if ((FEED_ELEMENT_NAME.equalsIgnoreCase(qName)) && (feedCount == 0)) {
			inFeed = true;
			this.response = new UpcResponseImpl();
			this.product = new UpcProductImpl();
			this.company = new UpcCompanyImpl();
			this.response.setProduct(this.product);
			this.response.setCompany(this.company);
		}
		if (inFeed) {
			if ((STATUS_ELEMENT_NAME.equalsIgnoreCase(qName)) && (statusCount == 0)) {
				statusCount++;
				inStatus = true;
			} else if ((PRODUCT_ELEMENT_NAME.equalsIgnoreCase(qName)) && (productCount == 0)) {
				productCount++;
				inProduct = true;
			} else if ((COMPANY_ELEMENT_NAME.equalsIgnoreCase(qName)) && (companyCount == 0)) {
				companyCount++;
				inCompany = true;
			} else if (inStatus || inProduct || inCompany ) {
				if (STATUS_ELEMENT_NAME.equalsIgnoreCase(qName)) {
					statusCount++;
				} else if (PRODUCT_ELEMENT_NAME.equalsIgnoreCase(qName)) {
					productCount++;
				} else if (COMPANY_ELEMENT_NAME.equalsIgnoreCase(qName)) {
					companyCount++;
				}
				currentName = qName;
			} else {
				String x = "x";
			}
		}
	}

	public void endElement(String uri, String localName, String qName) throws SAXException {

		if (FEED_ELEMENT_NAME.equalsIgnoreCase(qName)) {
			feedCount--;
			if (feedCount == 0) {
				inFeed = false;
				currentName = "";
			}
		} else if (STATUS_ELEMENT_NAME.equalsIgnoreCase(qName)) {
			statusCount--;
			if (statusCount == 0) {
				inStatus = false;
				currentName = "";
			}
		} else if (PRODUCT_ELEMENT_NAME.equalsIgnoreCase(qName)) {
			productCount--;
			if (productCount == 0) {
				inProduct = false;
				currentName = "";
			}
		} else if (COMPANY_ELEMENT_NAME.equalsIgnoreCase(qName)) {
			companyCount--;
			if (companyCount == 0) {
				inCompany = false;
				currentName = "";
			}
		}
	}

	public void characters(char ch[], int start, int length) throws SAXException {
		String value = new String(ch, start, length);
		if ("\n".equalsIgnoreCase(value)) {
			return;
		}
		if (inStatus) {
			if (VERSION_ELEMENT_NAME.equalsIgnoreCase(currentName)) {
				response.setResponseVersion(Integer.valueOf(new String(ch, start, length)));
			} else if (CODE_ELEMENT_NAME.equalsIgnoreCase(currentName)) {
				response.setReturnCode(Integer.valueOf(new String(ch, start, length)));
			} else if (MESSAGE_ELEMENT_NAME.equalsIgnoreCase(currentName)) {
				response.setRequestType(new String(ch, start, length));
			} else if (FIND_ELEMENT_NAME.equalsIgnoreCase(currentName)) {
				response.setUpc(new String(ch, start, length));
			}
		} else if (inProduct) {
			if (MODIFIED_ELEMENT_NAME.equalsIgnoreCase(currentName)) {
				try {
					product.setLastModified(dateFormatter.parse(new String(ch, start, length)));
				} catch(ParseException e) {
					//Do nothing...
				}
			} else if (EAN13_ELEMENT_NAME.equalsIgnoreCase(currentName)) {
				product.setEan13(new String(ch, start, length));
			} else if (UPCA_ELEMENT_NAME.equalsIgnoreCase(currentName)) {
				product.setUpcA(new String(ch, start, length));
			} else if (UPCE_ELEMENT_NAME.equalsIgnoreCase(currentName)) {
				product.setUpcE(new String(ch, start, length));
			} else if (ISBN_ELEMENT_NAME.equalsIgnoreCase(currentName)) {
				product.setIsbn(new String(ch, start, length));
			} else if (PRODUCT_ELEMENT_NAME.equalsIgnoreCase(currentName)) {
				product.setName(new String(ch, start, length));
			} else if (DESCRIPTION_ELEMENT_NAME.equalsIgnoreCase(currentName)) {
				product.setDescription(new String(ch, start, length));
			} else if (LONG_DESCRIPTION_ELEMENT_NAME.equalsIgnoreCase(currentName)) {
				product.setLongDescription(Boolean.valueOf(new String(ch, start, length)));
			} else if (CATEGORY_NUMBER_ELEMENT_NAME.equalsIgnoreCase(currentName)) {
				product.setCategoryCode(new String(ch, start, length));
			} else if (CATEGORY_TEXT_ELEMENT_NAME.equalsIgnoreCase(currentName)) {
				product.setCategoryName(new String(ch, start, length));
			} else if (URL_ELEMENT_NAME.equalsIgnoreCase(currentName)) {
				product.setCompanyUrl(new String(ch, start, length));
			} else if (IMAGE_ELEMENT_NAME.equalsIgnoreCase(currentName)) {
				product.setProductImageUrl(new String(ch, start, length));
				if (autoRetrieveGraphics) {
					product.setProductImage(UpcUtility.retrieveImage(product.getProductImageUrl()));
				}
			} else if (BARCODE_ELEMENT_NAME.equalsIgnoreCase(currentName)) {
				product.setBarcodeUrl(new String(ch, start, length));
				if (autoRetrieveGraphics) {
					product.setBarcodeImage(UpcUtility.retrieveImage(product.getProductImageUrl()));
				}
			} else if (LOCKED_ELEMENT_NAME.equalsIgnoreCase(currentName)) {
				product.setLocked(Boolean.valueOf(new String(ch, start, length)));
			}
		} else if (inCompany) {
			if (NAME_ELEMENT_NAME.equalsIgnoreCase(currentName)) {
				company.setName(new String(ch, start, length));
			} else if (LOGO_ELEMENT_NAME.equalsIgnoreCase(currentName)) {
				company.setLogoUrl(new String(ch, start, length));
				if (autoRetrieveGraphics) {
					company.setLogoImage(UpcUtility.retrieveImage(company.getLogoUrl()));
				}
			} else if (URL_ELEMENT_NAME.equalsIgnoreCase(currentName)) {
				company.setUrl(new String(ch, start, length));
			} else if (ADDRESS_ELEMENT_NAME.equalsIgnoreCase(currentName)) {
				company.setAddress(new String(ch, start, length));
			} else if (PHONE_ELEMENT_NAME.equalsIgnoreCase(currentName)) {
				company.setTelephone(new String(ch, start, length));
			} else if (LOCKED_ELEMENT_NAME.equalsIgnoreCase(currentName)) {
				company.setLocked(Boolean.valueOf(new String(ch, start, length)));
			}
		}
	}

}
