package alrosine.json;

import alrosine.json.exception.JsonInvalidFormatException;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

public class JsonGenerator {

	private enum ValueType {
		  ArrayBegin
		, ArrayEnd
		, ObjectBegin
		, ObjectEnd
		, Value
	}

	private static final String COMMA_DELIMITER_VALUE = ",";
	private static final String OPEN_ARRAY_VALUE = "[";
	private static final String CLOSE_ARRAY_VALUE = "]";
	private static final String QUOTE_VALUE = "\"";
	private static final String DECIMAL_VALUE = ".";
	private static final String EXPONENTIAL_VALUE = "e";
	private static final String EMPTY_ARRAY_VALUE = OPEN_ARRAY_VALUE + CLOSE_ARRAY_VALUE;
	private static final String NULL_VALUE = "null";
	private static final String TRUE_VALUE = "true";
	private static final String FALSE_VALUE = "false";

	private String jsonString;
	private byte[] jsonBytes;
	private int currentPosition = 0;
	private JsonObject currentJsonObject;
	private JsonObject lastJsonObject;

	private List<JsonObject> pushList = new ArrayList<JsonObject>();

	private JsonGenerator(String jsonString) {
		this.setJsonString(jsonString);
	}

	private JsonGenerator(byte[] jsonBytes) {
		this.setJsonString(jsonBytes);
	}

	public static JsonObject generate(String jsonString) {
		JsonGenerator gen = new JsonGenerator(jsonString);
		gen.parseObject();
		return gen.getJsonObject();
	}

	public static JsonObject generate(byte[] jsonBytes) {
		JsonGenerator gen = new JsonGenerator(jsonBytes);
		gen.parseObject();
		return gen.getJsonObject();
	}

	void parseObject() {
		if (!ValueType.ObjectEnd.equals(positionNextObject())) {
			if (this.currentPosition > -1) {
				push();
				ValueType valueType = null;
				while((this.currentPosition > -1) && (!ValueType.ObjectEnd.equals(valueType))) {
					String objectName = getObjectName();
					valueType = positionAfterNextObjectSeparator();
					Object object = null;
					switch(valueType) {
						case ArrayBegin:
							object = parseArray();
							break;
						case ObjectBegin:
							parseObject();
							object = this.lastJsonObject;
							break;
						case Value:
							object = parseValue();
							break;
					}
					this.currentJsonObject.addObject(objectName, object);
					valueType = positionNextObject();
				}
				pop();
			}
		}
	}

	List<Object> parseArray() {
		if (this.jsonBytes[this.currentPosition] == 32) {
			this.currentPosition = findNextNonWhitespace(this.currentPosition);
		}
		ValueType valueType = positionAfterNextArraySeparator();
		List<Object> objects = new ArrayList<Object>();
		while((this.currentPosition > -1) && (!ValueType.ArrayEnd.equals(valueType))) {
			Object object = null;
			switch(valueType) {
				case ArrayBegin:
					object = parseArray();
					break;
				case ObjectBegin:
					parseObject();
					break;
				case Value:
					object = parseValue();
					break;
			}
			objects.add(object);
			valueType = positionNextObject();
		}
		return objects;
	}

	Object parseValue() {
		int idx = this.currentPosition;
		if (this.jsonBytes[this.currentPosition] == 32) {
			idx = findNextNonWhitespace(this.currentPosition);
		}
		if (this.jsonBytes[idx] == 34) {
			//Opening Quote
			int closingQuote = findClosingQuote(idx);
			String stringValue;
			if (idx == (closingQuote - 1)) {
				stringValue = new String();
			} else {
				stringValue = this.jsonString.substring(idx + 1, closingQuote);
			}
			this.currentPosition = findValueCloser(closingQuote);
			return stringValue;
		} else {
			int valueCloser = findValueCloser(idx);
			if (idx == (valueCloser - 1)) {
				return null;
			}
			String stringValue = this.jsonString.substring(idx, valueCloser).trim();
			if (NULL_VALUE.equalsIgnoreCase(stringValue)) {
				this.currentPosition = valueCloser;
				return null;
			} else if (TRUE_VALUE.equalsIgnoreCase(stringValue)) {
				this.currentPosition = valueCloser;
				return Boolean.TRUE;
			} else if (FALSE_VALUE.equalsIgnoreCase(stringValue)) {
				this.currentPosition = valueCloser;
				return Boolean.FALSE;
			} else {
				this.currentPosition = valueCloser;
				return formatNumber(stringValue);
			}
		}
	}

	String getObjectName() throws JsonInvalidFormatException{
		int startPos = this.currentPosition;
		if (this.jsonBytes[this.currentPosition] == 32) {
			startPos = findNextNonWhitespace(this.currentPosition);
		}
		for(int idx = startPos;idx < this.jsonBytes.length;idx++) {
			if (this.jsonBytes[idx] == 34) {
				//Opening Quote
				int closingQuote = findClosingQuote(idx);
				if (idx == (closingQuote - 1)) {
					throw new JsonInvalidFormatException("Object name cannot be an empty string", this.currentPosition);
				}
				String stringValue = this.jsonString.substring(idx + 1, closingQuote);
				this.currentPosition = findValueCloser(closingQuote);
				return stringValue;
			}
		}
		return null;
	}

	int findClosingQuote(int openingQuote) {
		boolean insideQuote = false;
		int pos = -1;
		if (this.jsonBytes[openingQuote] == 34) {
			insideQuote = true;
		}
		if (insideQuote) {
			for(int idx = openingQuote + 1;idx < this.jsonBytes.length; idx++) {
				switch (this.jsonBytes[idx]) {
					case 92:
						//Escape character
						if (this.jsonBytes[idx + 1] == 117) {
							//Skip Unicode characters
							idx += 5;
						} else {
							//Skip control characters
							idx += 1;
						}
						continue;
					case 34:
						pos = idx;
				}
				if (pos != -1) {
					break;
				}
			}
		}
		return pos;
	}

	int findNextNonWhitespace(int currentPos) {
		for(int idx = currentPos;idx < this.jsonBytes.length; idx++) {
			switch (this.jsonBytes[idx]) {
				case 32:  //comma
					continue;
				default:
					return idx;
			}
		}
		throw new JsonInvalidFormatException("Unable to find next non-whitespace from position", currentPos);
	}

	int findValueCloser(int startPos) throws JsonInvalidFormatException {
		for(int idx = startPos;idx < this.jsonBytes.length; idx++) {
			switch (this.jsonBytes[idx]) {
				case 44:  //Comma
				case 58:  //Semicolon
				case 93:  //Closing bracket
				case 125: //Closing brace
					return idx;
			}
		}
		throw new JsonInvalidFormatException("Unable to find value closer starting from position", startPos);
	}

	ValueType positionNextObject() {
		for(int idx = this.currentPosition;idx < this.jsonBytes.length; idx++) {
			switch (this.jsonBytes[idx]) {
				case 44:  //Comma
					this.currentPosition = idx + 1;
					return getNextValueType();
				case 93:  //Closing bracket
					this.currentPosition = idx + 1;
					return ValueType.ArrayEnd;
				case 123:  //Opening brace
					this.currentPosition = idx + 1;
					return ValueType.ObjectBegin;
				case 125:  //Closing brace
					this.currentPosition = idx + 1;
					return ValueType.ObjectEnd;
			}
		}
		this.currentPosition = -1;
		return getNextValueType();
	}

	ValueType positionAfterNextObjectSeparator() throws JsonInvalidFormatException {
		for(int idx = this.currentPosition;idx < this.jsonBytes.length; idx++) {
			switch (this.jsonBytes[idx]) {
				case 58:  //ObjectBegin name/value separator
					this.currentPosition = findNextNonWhitespace(idx + 1);
					return getNextValueType();
			}
		}
		throw new JsonInvalidFormatException("Unable to find object separator ",this.currentPosition);
	}

	ValueType positionAfterNextArraySeparator() throws JsonInvalidFormatException {
		for(int idx = this.currentPosition;idx < this.jsonBytes.length; idx++) {
			switch (this.jsonBytes[idx]) {
				case 91:  //ObjectBegin value array
					this.currentPosition = findNextNonWhitespace(idx + 1);
					return getNextValueType();
			}
		}
		throw new JsonInvalidFormatException("Unable to find array separator ",this.currentPosition);
	}

	ValueType getNextValueType() {
		int idx = this.currentPosition;
		if (this.jsonBytes[this.currentPosition] == 32) {
			idx = findNextNonWhitespace(this.currentPosition);
		}
		switch (this.jsonBytes[idx]) {
			case 91:  //Opening bracket
				return ValueType.ArrayBegin;
			case 93:  //Closing bracket
				return ValueType.ArrayEnd;
			case 123:  //Opening brace
				return ValueType.ObjectBegin;
			case 125:  //Closing brace
				return ValueType.ObjectEnd;
		}
		return ValueType.Value;
	}

	String buildArray(Collection collection) {
		if ((collection == null) || (collection.size() == 0)) {
			return EMPTY_ARRAY_VALUE;
		}
		List<String> objects = new ArrayList<String>(collection.size());
		int size = 0;
		int idx = 0;
		for(Object obj : collection) {
			idx++;
			if (obj == null) {
				objects.set(idx, NULL_VALUE);
				size += 4;
			} else if (obj instanceof Boolean) {
				if (((Boolean)obj)) {
					objects.set(idx, TRUE_VALUE);
					size += 4;
				} else {
					objects.set(idx, FALSE_VALUE);
					size += 5;
				}
			} else if (obj instanceof Collection) {
				String objString = buildArray((Collection) obj);
				objects.set(idx, TRUE_VALUE);
				size += objString.length();
			} else if (obj instanceof Number) {
				String objString = obj.toString();
				objects.set(idx, objString);
				size += objString.length();
			} else {
				String objString = QUOTE_VALUE + obj.toString() + QUOTE_VALUE;
				objects.set(idx, objString);
				size += objString.length() + 2;
			}
			idx++;
		}
		return buildList(objects, size);
	}

	Number formatNumber(String value) throws JsonInvalidFormatException {
		try {
			if (value.toLowerCase().contains(EXPONENTIAL_VALUE)) {
				//Number values with an Exponent indicator are stored as a BigDecimal to indicate the
				return new BigDecimal(value);
			} else if (value.toLowerCase().contains(DECIMAL_VALUE)) {
				return Double.valueOf(value);
			} else {
				return Long.valueOf(value);
			}
		} catch(NumberFormatException e) {
			throw new JsonInvalidFormatException("Number value is not valid ", this.currentPosition);
		}
	}

	private String buildList(List<String> values, int size) {
		StringBuilder builder = new StringBuilder(size + (values.size() * 3) + 2);
		builder.append(OPEN_ARRAY_VALUE);
		String delimiter = "";
		for(String value : values) {
			builder.append(delimiter).append(value);
			delimiter = COMMA_DELIMITER_VALUE;
		}
		builder.append(CLOSE_ARRAY_VALUE);
		return builder.toString();
	}

	JsonObject getJsonObject() {
		return currentJsonObject;
	}

	void setJsonString(final String jsonString) {
		if (jsonString == null) {
			this.jsonString = null;
			this.jsonBytes = null;
		} else {
			this.jsonString = jsonString.trim();
			this.jsonBytes = this.jsonString.getBytes();
		}
		this.currentPosition = 0;
	}

	void setJsonString(final byte[] jsonBytes) {
		if (jsonBytes == null) {
			this.jsonString = null;
			this.jsonBytes = null;
		} else {
			this.jsonBytes = jsonBytes;
			this.jsonString = new String(this.jsonBytes).trim();
		}
		this.currentPosition = 0;
	}

	private void push() {
		this.pushList.add(new JsonObject());
		this.currentJsonObject = this.pushList.get(this.pushList.size() - 1);
	}

	private void pop() {
		this.lastJsonObject = this.currentJsonObject;
		this.pushList.remove(this.pushList.size() - 1);
		if (this.pushList.size() > 0) {
			this.currentJsonObject = this.pushList.get(this.pushList.size() - 1);
		}
	}

}
