package alrosine.json;

import alrosine.json.exception.JsonDuplicateNameException;
import alrosine.json.exception.JsonInvalidFormatException;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Since Oracle/Sun is currently in the process of building an implementation of the JSON format this is a personal implementation used to process
 * the JSON response from the eandata.com web site.
 *
 * This JsonObject represents the name/value pairs provided for by the JSON format.
 */
public class JsonObject {

	private static final String START_JSON_OBJECT = "{";
	private static final String END_JSON_OBJECT = "}";
	private static final String START_JSON_ARRAY = "[";
	private static final String END_JSON_ARRAY = "]";
	private static final String OBJECT_SEPARATOR_OBJECT = ":";
	private static final String VALUE_SEPARATOR_OBJECT = ",";
	private static final String STRING_DOUBLE_QUOTE_OBJECT = "\"";
	private static final String NULL_VALUE = "null";


	private Map<Integer, String> positions = new HashMap<Integer, String>();
	private Map<String, Integer> names = new HashMap<String, Integer>();
	private List<Object> objects = new ArrayList<Object>();

	/**
	 * The names of all values parsed. These names are in the same order that they were
	 * acquired from the original parsed JSON input.
	 *
	 * @return Set<String> - The names of the objects for the specific JsonObject being accessed.
	 */
	public Set<String> getNames() {
		return this.names.keySet();
	}

	/**
	 * Used to retrieve a specific value object for a provided name.
	 *
	 * @param name - The name of the desired value object.
	 *
	 * @return Object - The specific value object for the name provided. Null if name not found.
	 */
	public Object getObject(String name) {
		Integer idx = this.names.get(name);
		if (idx == null) {
			return null;
		}
		return this.objects.get(idx);
	}

	/**
	 * Used to add a name and value object to the JsonObject being built.
	 *
	 * @param name - The String name of the object being stored.
	 *
	 * @param object - The value object for the name being stored.
	 *
	 * @throws JsonDuplicateNameException - Thrown if the name of the object being stored has already been added to this JsonObject.
	 */
	public void addObject(String name, Object object) throws JsonDuplicateNameException {
		if (this.names.containsKey(name)) {
			throw new JsonDuplicateNameException(name);
		}
		Integer idx = this.objects.size();
		this.names.put(name, idx);
		this.positions.put(idx, name);
		this.objects.add(object);
	}

	/**
	 * The value objects for this JsonObject. These objects are in the same order that they were
	 * acquired from the original parsed JSON input.
	 */
	public List<Object> getObjects() {
		return objects;
	}

	/**
 	 * @return String - The original JSON String that was used to build this JsonObject.
	 *
	 * @throws JsonInvalidFormatException - Thrown when an unknown/unexpected data type is found.
	 */
	public String toString() throws JsonInvalidFormatException {
		StringBuilder builder = new StringBuilder(START_JSON_OBJECT);
		String separator = "";
		for(Map.Entry<Integer, String> entry : this.positions.entrySet()) {
			builder.append(separator);
			builder.append(STRING_DOUBLE_QUOTE_OBJECT)
					.append(entry.getValue())
					.append(STRING_DOUBLE_QUOTE_OBJECT)
					.append(OBJECT_SEPARATOR_OBJECT)
			;
			Object obj = this.objects.get(entry.getKey());
			if (obj == null) {
				builder.append(NULL_VALUE);
			} else if (obj instanceof String) {
				addString(builder, (String) obj);
			} else if (obj instanceof Boolean) {
				builder.append(obj.toString());
			} else if (obj instanceof List) {
				addList(builder, (List) obj);
			} else if (obj instanceof JsonObject) {
				builder.append(obj.toString());
			} else if (obj instanceof Number) {
				builder.append(obj.toString());
			} else {
				throw new JsonInvalidFormatException("Unknown object type found ", obj.getClass());
			}
			separator = VALUE_SEPARATOR_OBJECT;
		}
		builder.append(END_JSON_OBJECT);
		return builder.toString();
	}

	/**
	 * Used to format and add a String value object to the output StringBuilder object.
	 *
	 * @param builder - The StringBuilder to which the String value object is to be added.
	 * @param value - The String value object to be quoted and added to the StringBuilder.
	 */
	private void addString(StringBuilder builder, String value) {
		builder.append(STRING_DOUBLE_QUOTE_OBJECT)
				.append(encodeString(value))
				.append(STRING_DOUBLE_QUOTE_OBJECT);
	}

	private String encodeString(String decodedValue) {
		//TODO: needs to be tested.
		String encodedValue = new String(decodedValue);
		for(Encoding cc : Encoding.values()) {
			int idx = encodedValue.indexOf(cc.getDecodedValue());
			while(idx > -1) {
				if (idx == 0) {
					encodedValue = cc.getEncodedValue() + encodedValue.substring(idx + 2);
				} else {
					encodedValue = encodedValue.substring(0, idx) + cc.getEncodedValue() + encodedValue.substring(idx + cc.getDecodedValue().length());
				}
				idx = encodedValue.indexOf(cc.getDecodedValue(), idx + cc.getEncodedValue().length());
			}
		}
		return encodedValue;
	}
	/**
	 * Used to format and add a List of value objects to the output StringBuilder object.
	 *
	 * @param builder - The StringBuilder to which the String value object is to be added.
	 * @param values - The List containing the value objects to be added.
	 */
	private void addList(StringBuilder builder, List values) {
		builder.append(START_JSON_ARRAY);
		String separator = "";
		if ((values != null) && (values.size() > 0)) {
			for(Object obj : values) {
				builder.append(separator);
				if (obj == null) {
					builder.append(NULL_VALUE);
				} else if (obj instanceof String) {
					addString(builder, (String) obj);
				} else if (obj instanceof Boolean) {
					builder.append(obj.toString());
				} else if (obj instanceof List) {
					addList(builder, (List) obj);
				} else if (obj instanceof JsonObject) {
					builder.append(obj.toString());
				} else {
					throw new JsonInvalidFormatException("Unknown object type found ", obj.getClass());
				}
				separator = VALUE_SEPARATOR_OBJECT;
			}
		}
		builder.append(END_JSON_ARRAY);
	}

}
